/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: auwith.c,v 1.2 2014/09/16 17:54:09 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <strings.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "auwith_int.h"

static int *mode = NULL;

void
/*****************************************************************************************************/
fase(auwith_opt *opt, FILE *fo) {                                                                /****
 *****************************************************************************************************/
    auwith_sfaf sfaf;
    u_int i, o;
    int f;

    if  (!mode && !(mode = calloc(opt->pidm->nout * opt->pidm->ninp, sizeof(int)))) {
        ifBLIN_QX0("Malloc failed #1");
        errno = ENOMEM;
        goto out;
    }
    fprintf(fo, "i\\o");
    for (o = 0; o < opt->pidm->nout; ++o) fprintf(fo, "%s%5u", o ? " " : "", o);
    fprintf(fo, "\n");
    for (i = 0; i < opt->pidm->ninp; ++i) {
        fprintf(fo, "%2u", i);
        for (o = 0; o < opt->pidm->nout; ++o) {
            if  (0 < mode[opt->pidm->nout * i + o]) --mode[opt->pidm->nout * i + o];
            if  (auwith_cella(opt->pidm, i, o)->t.doo) mode[opt->pidm->nout * i + o] = 30;
            auwith_cella(opt->pidm, i, o)->t.doo = 0;
            if  (!!mode[opt->pidm->nout * i + o]) {
                bcopy(&auwith_cella(opt->pidm, i, o)->t, &sfaf, sizeof(auwith_sfaf));
                f = auwith_sfafer(&sfaf);
            } else {
                f = auwith_cella(opt->pidm, i, o)->f.faser;
            }
            if  (!f) {
                fprintf(fo, "      ");
            } else if (opt->flags & AUWITH_FAEX) {
                fprintf(fo, "%6d", f);
            } else {
                fprintf(fo, "%6d", f * 100 / AUWITH_FASMUL);
        }   }
        fprintf(fo, "\n");
    }
out:
    fflush(fo);
}

static void
/*****************************************************************************************************/
usage(int ex) {                                                                                  /****
 *****************************************************************************************************/
    fprintf( stderr
           , "auwith " VERS " @BABOLO " DATE "\n"
             "Usage: auwith [-options]\n"
             "   -a dev          DSP to work with, default %s\n"
             "   -d in out       deconfig matrix cell [in, out]\n"
             "   -D              dialog mode\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -f file         auwithd's control file\n"
             "   -F num          filter for -T\n"
             "   -I num          stream level for -i num\n"
             "   -l in out level config matrix cell [in, out] with level\n"
             "   -m dev          mixer to work with, default %s\n"
             "   -M              look at matrix state\n"
             "   -MM             master of peek volumes\n"
             "   -O num          stream level for -o num\n"
             "   -P              phase indication in 1/100\n"
             "   -PP             phase indication in internal units (1/%u)\n"
             "   -q              silent\n"
             "   -u              us instead of ns for -T\n"
             "   -v[v[v...]]     verbose\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -= mode         internal diagnostic\n"
             "   -?              this text\n"
             "  priority -=0123, -ld, -T, -=45, -IO, -M, -P, -D\n"
           , AUWITH_DEV
           , AUWITH_MIX
           , AUWITH_FASMUL
           )
    ;
    exit(ex);
}

static void
/*****************************************************************************************************/
levout(FILE *fo, auwith_vol *vol) {                                                              /****
 *****************************************************************************************************/
        int v, mi, ma, m;

        mi = vol->min;
        ma = vol->max;
        vol->min = 0;
        vol->max = 0;
        if  (mi < 0) mi = -mi;
        if  (ma < 0) ma = -ma;
        v = (mi > ma) ? mi : ma;
        for (m = 0; v; ++m) v >>= 1;
        fprintf(fo, "%d\n", m);
        fflush(fo);
}

static int
/*****************************************************************************************************/
opid(auwith_opt *opt) {                                                                          /****
 *****************************************************************************************************/
    int ex = EX_OK;
    struct stat sb;
    ssize_t u;

    if  (0 > (opt->pidd = open(opt->pidf, O_RDWR, 0600))) {
        ifBLIN_QW0("open pid");
        EXOUT(EX_IOERR);
    }
    if  (0 > fstat(opt->pidd, &sb)) {
        ifBLIN_QW0("fstat");
        EXOUT(EX_IOERR);
    }
    if  (!S_ISREG(sb.st_mode)) {
        ifBLIN_QX0("%s not reg file", opt->pidf);
        ERROUT(EX_SOFTWARE, EFTYPE);
    }
    u = sizeof(auwith_pid);
    if  (sb.st_size < u) {
        ifBLIN_QX0("%s too short %d < %d", opt->pidf, sb.st_size, u);
        ERROUT(EX_SOFTWARE, EFTYPE);
    }
    opt->pidm = mmap(NULL, sb.st_size, PROT_READ | PROT_WRITE, MAP_NOSYNC | MAP_SHARED, opt->pidd, 0);
    if  (MAP_FAILED == opt->pidm) {
        ifBLIN_QW0("mmap");
        EXOUT(EX_IOERR);
    }
    close(opt->pidd);
    u = sizeof(auwith_cell) * opt->pidm->ninp * opt->pidm->nout + sizeof(auwith_pid);
    u += sizeof(auwith_vol) * (opt->pidm->ninp + opt->pidm->nout);
    if  (sb.st_size < u) {
        ifBLIN_QX0("%s too short %d < %d", opt->pidf, sb.st_size, u);
        ERROUT(EX_SOFTWARE, EFTYPE);
    }
    opt->inpvol = (auwith_vol*)&opt->pidm->matrix[opt->pidm->nout * opt->pidm->ninp];
    opt->outvol = &opt->inpvol[opt->pidm->ninp];
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
}

static int
/*****************************************************************************************************/
parsin(auwith_opt *opt, int argc, char *argv[]) {                                                /****
 *****************************************************************************************************/
    char *flar = "a:d::Df:F:I:l:::m:MO:Pqt:T:uv=:?";
    int c, i, o, ex = EX_OK;
    struct timespec tn;

    ifBLIN_QX3("+");
    if  (!(opt->bos = babolo_openopts((opt->flags & BLIN_MASK) | Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("openopts");
        EXOUT(EX_SOFTWARE);
    }
    if  (!!babolo_setopts(opt->bos, (opt->flags & BLIN_MASK) | Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("setopts");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_USAGE;
    while (!!(c = babolo_getopts(opt->bos))) {
        switch (c) {
        case 'a': if  (!(opt->env.dev = auwith_todev(babolo_getoptsarg(opt->bos), AUWITH_DEV)))
                      goto out;
                ; break;
        case 'd': if  (!opt->pidm) {
                      ifBLIN_QX0("Use -f option before -d");
                      ERROUT(EX_USAGE, EPROTO);
                  }
                ; i = babolo_getoptunum(opt->bos, 0);
                ; o = babolo_getoptunum(opt->bos, 0);
                ; auwith_cella(opt->pidm, i, o)->flags &= ~AUWITH_USED;
                ; break;
        case 'D': opt->flags |= AUWITH_DIAL;
                ; break;
        case 'f': opt->pidf = babolo_getoptsarg(opt->bos);
                ; if  (!!(ex = opid(opt))) goto out
                ; break;
        case 'F': opt->Tfilter = babolo_getoptunum(opt->bos, 0);
                ; opt->flags |= AUWITH_FILT;
                ; break;
        case 'I': opt->sock = babolo_getoptunum(opt->bos, 0);
                ; opt->flags &= ~AUWITH_OLEV;
                ; opt->flags |= AUWITH_ILEV;
                ; break;
        case 'l': if  (!opt->pidm) {
                      ifBLIN_QX0("Use -f option before -l");
                      ERROUT(EX_USAGE, EPROTO);
                  }
                ; i = babolo_getoptunum(opt->bos, 0);
                ; o = babolo_getoptunum(opt->bos, 0);
                ; auwith_cella(opt->pidm, i, o)->flags
                      = (babolo_getoptunum(opt->bos, 0) & AUWITH_LEVEL)
                      | (auwith_cella(opt->pidm, i, o)->flags & AUWITH_ALGO)
                      | AUWITH_USED
                  ;
                ; break;
        case 'm': if  (!(opt->env.mix = auwith_todev(babolo_getoptsarg(opt->bos), AUWITH_MIX)))
                      goto out;
                ; break;
        case 'M': if  (opt->flags & AUWITH_LOOK) opt->flags |= AUWITH_MAST;
                ; opt->flags |= AUWITH_LOOK;
                ; break;
        case 'O': opt->sock = babolo_getoptunum(opt->bos, 0);
                ; opt->flags &= ~AUWITH_ILEV;
                ; opt->flags |= AUWITH_OLEV;
                ; break;
        case 'P': if  (opt->flags & AUWITH_FASE) opt->flags |= AUWITH_FAEX;
                ; opt->flags |= AUWITH_FASE;
                ; break;
        case 'q': BLIN_QUIET(opt->flags);
                ; break;
        case 't': if  (!!(ex = auwith_totime(babolo_getoptsarg(opt->bos), &tn))) goto out;
                ; opt->tr = (u_int64_t)tn.tv_sec * 1000000000ULL + tn.tv_nsec;
                ; break;
        case 'T': if  (!(opt->T = babolo_getoptsarg(opt->bos))) goto out;
                ; break;
        case 'u': opt->flags |= AUWITH_uSEC;
                ; break;
        case 'v': BLIN_VERBOSE(opt->flags);
                ; break;
        case '=': switch((opt->tt = babolo_getoptsarg(opt->bos)[0])) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(opt->bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(opt->bos);         exit(EX_OK);
                  }
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c",opt->bos->c);
                ; usage(EX_USAGE);
    }   }
    ex = EX_OK;
    opt->env.flags = opt->flags;
out:
    ifBLIN_QX3("- %d", ex);
    return(ex);
}

int
/*****************************************************************************************************
 *****************************************************************************************************/
main(int argc, char *argv[]) {                                                                   /****
 *****************************************************************************************************
 *****************************************************************************************************/
    auwith_opt *opt = NULL;
    struct timespec tn;
    int ex = EX_OK;
    u_int j;

    blin_ctl(BLIN_CTL_FSET, BLIN_BIT0);
//  blin_ctl(BLIN_CTL_FSET, BLIN_MASK);
    if  (!(opt = calloc(1, sizeof(auwith_opt)))) {
        ifBLIN_QX0("Malloc failed #1");
        ERROUT(EX_OSERR, ENOMEM);
    }
    opt->flags = blin_internal_flags & BLIN_MASK ;
    opt->env.mix = AUWITH_MIX;
    opt->env.dev = AUWITH_DEV;
    if  (!!(ex = parsin(opt, argc, argv))) goto out;
    tn.tv_sec = opt->tr / 1000000000ULL;
    tn.tv_nsec = opt->tr % 1000000000ULL;
    if  (!!opt->T) {
        auwith_time time;
        int t, l;
        struct {
            char        t[4];
            int32_t     c;
            auwith_time time;
        } mark;

        if  (0 > (t = open(opt->T, O_RDONLY))) {
            ifBLIN_QW0("open T=%s~", opt->T);
            EXOUT(EX_IOERR);
        }
        for (time = 0;;) {
            l = 0;
            if  (0 > read(t, &l, 1)) {
                ifBLIN_QW0("read Tl");
                EXOUT(EX_IOERR);
            }
            switch (l) {
            case 0:
                goto out;
            case 13:
                if  (0 > read(t, ((char*)&mark) + 4, 12)) {
                    ifBLIN_QW0("read Tl");
                    EXOUT(EX_IOERR);
                }
                if  (!(opt->flags & AUWITH_FILT) || (mark.c == opt->Tfilter)) {
                    if  (opt->flags & AUWITH_uSEC) {
                        printf("%11d %19llu %19llu\n", mark.c, (mark.time - time) / 1000, time / 1000);
                    } else {
                        printf("%11d %22llu %22llu\n", mark.c, mark.time - time, time);
                    }
                    time = mark.time;
                }
                break;
            case 16:
                if  (0 > read(t, ((char*)&mark) + 1, 15)) {
                    ifBLIN_QW0("read T");
                    EXOUT(EX_IOERR);
                }
                if  (opt->flags & AUWITH_uSEC) {
                     printf("----------- %19llu\n", mark.time / 1000);
                } else {
                     printf("----------- %22llu\n", mark.time);
                }
                time = mark.time;
                break;
            default:
                printf("???????????? %d\n", l);
        }   }
    } else if (!opt->pidm) {
        if  (0 > (opt->env.aumix = open(opt->env.mix, O_RDWR))) {
            ifBLIN_QW0("open audio mixer");
            EXOUT(EX_IOERR);
        }
        if  (0 > (opt->env.audev = open(opt->env.dev, O_RDWR))) {
            ifBLIN_QW0("open audio dev");
            EXOUT(EX_IOERR);
        }
        switch (opt->tt) {
        case '4': auwith_show(&opt->env, stdout, 15);
                ; fprintf(stdout, "\n");
                ; exit(EX_OK);
        case '5': auwith_ana(&opt->env, stdout);
                ; exit(EX_OK);
        }
    } else if (opt->flags & AUWITH_ILEV) {
        while (1) {
            levout(stdout, &opt->inpvol[opt->sock]);
            if  (!opt->tr) break;
            nanosleep(&tn, NULL);
        }
    } else if (opt->flags & AUWITH_OLEV) {
        while (1) {
            levout(stdout, &opt->outvol[opt->sock]);
            if  (!opt->tr) break;
            nanosleep(&tn, NULL);
        }
    } else if (opt->flags & AUWITH_LOOK) {
        while (1) {
            if  (!!opt->tr) printf("%c%c%c%c%c%c%c", 0x1B, 0x5B, 0x48, 0x1B, 0x5B, 0x32, 0x4A);
            auwith_matrix(opt->pidm, stdout);
            for (j = 0; j < opt->pidm->ninp; ++j) opt->inpvol[j].min = opt->inpvol[j].max = 0;
            for (j = 0; j < opt->pidm->nout; ++j) opt->outvol[j].min = opt->outvol[j].max = 0;
            if  (!opt->tr) break;
            nanosleep(&tn, NULL);
        }
    } else if (opt->flags & AUWITH_FASE) {
        while (1) {
            if  (!!opt->tr) printf("%c%c%c%c%c%c%c", 0x1B, 0x5B, 0x48, 0x1B, 0x5B, 0x32, 0x4A);
            fase(opt, stdout);
            if  (!opt->tr) break;
            nanosleep(&tn, NULL);
        }
    } else {
        
    }
out:
    exit(ex);
}
