/*-
 * Copyright (C)2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: sine.c,v 1.1 2014/08/14 09:34:37 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3
#define Bpars_COMPAT 3

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "auwith_int.h"
                        /*****************************/
typedef struct {        /********* auwith_ent_sine ***/
    auwith_ent      common;
    u_int16_t       ampli;
    u_int32_t       freq;
    u_int32_t       phase;
} auwith_ent_sine;
                        /*****************************/
typedef struct {        /********* auwith_io_sine ****/
    auwith_io       common;
} auwith_io_sine;

static void
/*****************************************************************************************************/
usage(void) {                                                                                    /****
 *****************************************************************************************************/
    fprintf( stderr
           , "auwithd sine V%u @BABOLO " DATE "\n"
             "Usage: auwithd [-options] sine [-options]\n"
             "   -a ampl         sine's amplitude, max 32767\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -f freq         sine's frequency in Hz\n"
             "   -i in           matrix's input number\n"
             "   -L              local phase count\n"
             "   -p phase        sine's phase in 2pi/65536'th (default 0)\n"
             "   -q              silent\n"
             "   -r freq         samples rate in Hz (default %u)\n"
             "   -s num          block size in samples (default %u)\n"
             "   -v[v[v...]]     verbose\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -= mode         internal diagnostic\n"
             "   -?              this text\n"
           , auwith_def_sine.mark.ver
           , AUWITH_RATE
           , AUWITH_BLOCK
           )
    ;
}

static int
/*****************************************************************************************************/
parsin(auwith_opd *opd) {                                                                        /****
 *****************************************************************************************************/
    char *flar = "a:f:i:Lp:qr:s:v=:?";
    auwith_ent_sine *ent;
    int c, ex = EX_OK;
    u_int16_t inpp;

    ifBLIN_QX3("+");
    if  (!opd) {
        ifBLIN_QX0("no opd");
        ERROUT(EX_DATAERR, EFAULT);
    }
#   define blin_internal_flags (opd->flags & BLIN_MASK)
    if  (!opd->bos) {
        ifBLIN_QX0("no bos");
        ERROUT(EX_DATAERR, EFAULT);
    }
    if  (!(ent = calloc(1, sizeof(auwith_ent_sine)))) {
        ifBLIN_QX0("Malloc failed #1");
        ERROUT(EX_OSERR, ENOMEM);
    }
    ent->common.flags = (opd->flags & ~AUWITH_CAP_MASK) | (auwith_def_sine.flags & AUWITH_CAP_MASK);
    ent->common.plug = &auwith_def_sine;
    ent->common.rate = AUWITH_RATE;
    ent->common.block = AUWITH_BLOCK;
    ent->common.bits = AUWITH_BITS;
    if  (!!babolo_chopts(opd->bos, 0, flar)) {
        ifBLIN_QW0("chopts");
        EXOUT(EX_SOFTWARE);
    }
    ex = EX_OK;
    if  (opd->flags & AUWITH_DUMPOP) babolo_dumpopts(opd->bos);
    inpp = 0;
    while (!!(c = babolo_getopts(opd->bos))) {
        switch (c) {
        case 'a': ent->ampli = babolo_getoptunum(opd->bos, 0);
                ; break;
        /**************************************************************************************/
        case 'b': ent->common.bits = babolo_getoptunum(opd->bos, 0);           /* not reached */
                ; break;                                                       /* not reached */
        /**************************************************************************************/
        case 'f': ent->freq = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'i': if  (inpp >= 1) {
                      ifBLIN_QX0("chan %u >= %u", inpp, 1);
                      EXOUT(EX_USAGE);
                  }
                ; if  (!(ent->common.in[inpp] = calloc(1, sizeof(auwith_io_sine)))) {
                      ifBLIN_QX0("Malloc failed #2");
                      ERROUT(EX_OSERR, ENOMEM);
                  }
                ; ent->common.in[inpp]->flags = ent->common.flags;
                ; ent->common.in[inpp]->ent = (auwith_ent*)ent;
                ; ent->common.in[inpp]->sock = babolo_getoptunum(opd->bos, 0);
                ; if  (ent->common.in[inpp]->sock >= opd->inpachsz) {
                      ifBLIN_QX0( "-i[%u] must be < %u, now %u"
                                , inpp, opd->inpachsz, ent->common.in[inpp]->sock
                                );
                      ERROUT(EX_USAGE, EINVAL);
                  }
                ; ++inpp;
                ; break;
        case 'L': ent->common.flags ^= AUWITH_CFAF;
                ; break;
        case 'p': ent->phase = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'q': BLIN_QUIET(ent->common.flags);
                ; break;
        case 'r': ent->common.rate = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 's': ent->common.block = babolo_getoptunum(opd->bos, 0);
                ; break;
        case 'v': BLIN_VERBOSE(ent->common.flags);
                ; break;
        case '=': switch(babolo_getoptsarg(opd->bos)[0]) {
                  case '0': blin_ctl(BLIN_CTL_DUMP);           exit(EX_OK);
                  case '1': babolo_optest(flar);               exit(EX_OK);
                  case '2': babolo_optext(opd->bos);           exit(EX_OK);
                  case '3': babolo_dumpopts(opd->bos);         exit(EX_OK);
                  }
                ; break;
        case '?': usage();
                ; break;
        default : ifBLIN_QX0("Illegal flag -%c", opd->bos->c);
                ; usage();
                ; ERROUT(EX_USAGE, EINVAL);
    }   }
    if  (!ent->common.bits || (ent->common.bits > 16)) {
        ifBLIN_QX0("-b must be >0 and <= 16, now %u", ent->common.bits);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!ent->common.rate || (ent->common.rate > 48000)) {
        ifBLIN_QX0("-r must be >0 and <= 48000, now %u", ent->common.rate);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!ent->common.block || (ent->common.block > 16384)) {
        ifBLIN_QX0("-s must be >0 and <=16384, now %u", ent->common.block);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!ent->freq || (ent->freq * 2 > ent->common.rate)) {
        ifBLIN_QX0("-f must be >0 and <= %u, now %u", ent->common.rate, ent->freq);
        ERROUT(EX_USAGE, EINVAL);
    }
    if  (!!(ex = auwith_plumb(opd, ent->common.in[0], 0))) {
        ifBLIN_QX0("auwith_plumbi");
        goto out;
    }
out:
    if  (!!ex && !!ent->common.in[0]) free(ent->common.in[0]);
    ifBLIN_QX3("- %d", ex);
    return(ex);
#   undef blin_internal_flags
}

static void
/*****************************************************************************************************/
deconf(auwith_opd *opd, auwith_io *io, FILE *fo) {                                               /****
 *****************************************************************************************************/
    auwith_io_sine *cf = (auwith_io_sine*)io;
    auwith_ent_sine *ent;

#   define blin_internal_flags (io->flags & BLIN_MASK)
    ent = (auwith_ent_sine*)cf->common.ent;
    fprintf( fo, "[%u] sine %uHz, %usamples, A %u/32768 x freq %uHz @ %u/65536 phase/2pi\n"
           , cf->common.sock, ent->common.rate, ent->common.block, ent->ampli, ent->freq, ent->phase
           );
    fprintf(fo, "    ");
    auwith_capand(opd->flags, fo, 0);
    fprintf(fo, " ");
    auwith_capand(io->ent->flags, fo, 1);
    fprintf(fo, " ");
    auwith_capand(io->flags, fo, 1);
    fprintf(fo, "\n");
#   undef blin_internal_flags
}

static auwith_buf *
/*****************************************************************************************************/
reed(auwith_opd *opd, auwith_cell *cell) {                                                                        /****
 *****************************************************************************************************/
    auwith_buf *buf = NULL;
    auwith_ent_sine *ent;
    auwith_vol *inpvol;
    auwith_hc *is;
    int32_t w;
    double v;
    int j;

#   define blin_internal_flags (cell->flags & BLIN_MASK)
    ifBLIN_QX3("+");
    is = auwith_cadd(opd, cell);
    ent = (auwith_ent_sine*)opd->inpach[cell->inp]->ent;
    if  (!(buf = auwith_creabuf(ent->common.block * AUWITH_MWBY))) {                              /*@*/
        ifBLIN_QW0("creabuf");
    } else {
        inpvol = opd->inpvol;
        for (j = 0; j < ent->common.block; ++j) {
            if  (!ent->freq) {
                buf->samples[j] = 0;
            } else {
                if  ((u_int64_t)is->offset > ((u_int64_t)ent->freq * (u_int64_t)ent->common.rate)) {
                    is->offset -= (u_int64_t)ent->freq * (u_int64_t)ent->common.rate;
                }
                v = (double)is->offset * (double)ent->freq / (double)ent->common.rate;
                auwith_Tmark(opd, 2);
                v = sin((v + (double)ent->phase / 65536.0) * 2 * M_PI) * (double)ent->ampli;
                auwith_Tmark(opd, 3);
                w = v;
                if  (w < -32767) w = -32767;
                if  (w > 32767) w = 32767;
                buf->samples[j] = w;
                if  (inpvol[cell->inp].min > w) inpvol[cell->inp].min = w;
                if  (inpvol[cell->inp].max < w) inpvol[cell->inp].max = w;
                ++is->offset;
    }   }   }
    ifBLIN_QX3("- %"BLIN_X, BLIN_I(buf));
    return(buf);                                                                                  /*@*/
#   undef blin_internal_flags
}

/*****************************************************************************************************
 *****************************************************************************************************/
auwith_plug auwith_def_sine =                                                                    /****
 *****************************************************************************************************
 *****************************************************************************************************/
{ "sine"
, AUWITH_MARK
, parsin
, deconf
, NULL
, NULL
, reed
, NULL
, AUWITH_CAP_INP
, 0
};
